/*
Very simple GL program using bare OpenGL libraries to
demonstrate texturing

Orion Sky Lawlor, olawlor@acm.org, 2006/10/12 (Public Domain)
*/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>

#include <GL/gl.h> /* GL Extensions Wrangler, http://glew.sourceforge.net/ */
#include <GL/glut.h> /* GL Utilities Toolkit, http://freeglut.sourceforge.net/ */

#include "ltga.h" /* Lev Povalahev's targa-reading library; http://www.levp.de/3d/files/ltga.zip */
#include "ltga.cpp"

void Exit(const char *where,const char *why) {
	fprintf (stderr, "FATALITY: OpenGL Error in %s: %s\n", where, why);
	abort();	
}

// Check GL for errors after the last command
void Check(const char *where) {
	GLenum e=glGetError();
	if (e==GL_NO_ERROR) return;
	const GLubyte *errString = gluErrorString(e);
	Exit(where, (const char *)errString);
}

/************* Object we'll be drawing ***********/
class mainObject_t {
public:
	mainObject_t();
	void draw(void);
};


/** Set up a 3D model, stored in a vertex buffer */
mainObject_t::mainObject_t()
{
	/* nothing to set up here... */
}



/** Read a texture object from this file */
void readTextureTGA(const char *fName) 
{
#if 0 /* Simple 2x2 hardcoded image: */
	enum {w=2}; enum{h=2};
	unsigned char colors[4*w*h]={
		0xff,0x00,0x00,0xff,  0x00,0xff,0x00,0xff,
		0x00,0x00,0xff,0xff,  0xff,0xff,0xff,0x7f,
	};
	glTexImage2D(GL_TEXTURE_2D,
		0, // highest-level mipmap
		GL_RGBA8, // internal format
		w,h, // size of texture in pixels
		0, // border
		GL_RGBA, // format of my pixels
		GL_UNSIGNED_BYTE, // data type is unsigned char
		colors);
	
	
	return;
#endif

	LTGA f(fName); //<- read from a file, using LTGA library
	
	/* Careful!  Not all .tga images have an alpha channel! */
	GLenum rgbOrder=GL_RGBA;
	if (f.GetImageType()==itRGB) rgbOrder=GL_RGB;
	unsigned char *src=f.GetPixels();
	int w=f.GetImageWidth(), h=f.GetImageHeight();
	unsigned char *scratch=new unsigned char[4*(w/2)*(h/2)];
	unsigned char *dest=scratch;
	for (int level=0;w>0;level++) {
	  glTexImage2D(GL_TEXTURE_2D, // target: 1D, 2D, 3D, cubemap, ...
		level, // mipmap level
		GL_RGBA8, // Internal format
		w,h, // width and height
		0, // border
		rgbOrder, // RGBA or RGB
		GL_UNSIGNED_BYTE, // datatype of pixels
		src // actual pixel data
	  );
	  // For mipmapping: shrink image down for next mipmap level.
	  //  FIXME: assumes images are square, and 4-color (RGBA).
	  // For a better, easier way, see gluBuild2DMipmaps (but be careful!  This is a *slow* function!)
	  // For a faster way, see ogl/fast_mipmaps.c...
	  int dw=w/2, dh=h/2;
	  if (dw==0) break;
	  /* Loop over output pixels */
	  for (int y=0;y<dh;y++)
	  for (int x=0;x<dw;x++) {
		/* Loop over input pixels, from bigger source image */
		for (int c=0;c<4;c++) 
		{
			int avg=0;
			for (int sy=2*y;sy<2*(y+1);sy++)
			for (int sx=2*x;sx<2*(x+1);sx++)
				avg+=src[4*(sx+sy*w)+c];
			dest[4*(x+y*dw)+c]=(avg+3)/4;
			// Color selected mipmap levels:
			if (level==0)
	  			dest[4*(x+y*dw)+c]=0xff; // white
			else if (level==1)
	  			dest[4*(x+y*dw)+c]=0x5f; // dark grey
		}
	  }
	  w=dw; h=dh;
	  std::swap(dest,src);
	}
	delete[] scratch;
}

/** Set the GUI-selected OpenGL texture wrap and filtering modes on the current texture */
void setTextureModes(void) {
	GLenum target=GL_TEXTURE_2D;
	glTexParameteri(target, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_BORDER); 
	glTexParameteri(target, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_BORDER);
	/* WARNING: GL_LINEAR_MIPMAP_LINEAR results in white if you haven't set up all the mipmap levels! */
	glTexParameteri(target, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
	glTexParameteri(target, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	Check("After setting texture parameters");}

/** Main drawing routine, called by main_display below */
void mainObject_t::draw(void)
{
	Check("start draw");
	
	readTextureTGA("simple.tga");
	setTextureModes();
	
	glDisable(GL_LIGHTING);
	glEnable(GL_TEXTURE_2D);
	Check("After setting up texture");
	
	/* Finally, at long last, draw some friggin' objects! */
	glBegin(GL_TRIANGLE_FAN);
	glColor4f(1,1,1,1);
	glTexCoord2f(0,0);
	glVertex4f(-1,-1,0,1);
	glTexCoord2f(0,1);
	glVertex4f(-1,+1,0,1);
	glTexCoord2f(1,1);
	glVertex4f(+1,+1,0,1);
	glTexCoord2f(1,0);
	glVertex4f(+1,-1,0,1);
	glEnd();
	Check("After drawing object");
	
	//glBindTexture(GL_TEXTURE_2D,0); /* back to default texturing */
	
	/* This line causes the world to repeatedly redraw itself, using 100% of the CPU. */
	glutPostRedisplay();
	Check("end draw");
}


/******** Program Initialization *********/
mainObject_t *mainObject=0; /**< Big object we're trying to draw */


/* GLUT calls this routine every time our window needs to be redrawn */
void main_display(void) {
	Check("Entering display routine");
	// Clear the color and depth buffers
	glClearColor(0.2,0.3,0.4,1.0); /* set clear color */
	glClearDepth(1.0); /* set clear depth */
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT); /* actually clear color and depth */
	
	// Set up a little error-prone OpenGL state:
	glDisable(GL_POLYGON_SMOOTH); /* smooth edges of polygons */
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
	glEnable(GL_DEPTH_TEST);
	glPointSize(10.0); /* diameter in pixels */
	glLineWidth(10.0); /* thickness in pixels */
	
	// Draw stuff
	Check("About to draw");
	mainObject->draw();
	Check("Done drawing");
	
	// Display the stuff drawn so far to the screen:
	glutSwapBuffers();
}

/* GLUT calls this routine every time our window changes size */
void main_reshape(int w, int h) {
	Check("Before glViewport");
	int win_w=w, win_h=h;
	glViewport(0, 0, (GLsizei)win_w, (GLsizei)win_h);
	Check("After glViewport");
}

int main(int argc, char** argv)
{
	glutInitWindowSize(800,600);
	glutInit(&argc,argv);

	// Ask for a back buffer, color buffer, and Z buffer in a window.
	int mode=GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH;
	glutInitDisplayMode (mode);
	int main_window = glutCreateWindow("GL Bare-metal Texturing Demo");
	
	// Ask GLUT to call us when we need to redraw
	glutDisplayFunc (main_display);
	glutReshapeFunc (main_reshape);
	
	// Make an object to draw
	mainObject=new mainObject_t;
	
	// Process events until user closes our window...
	glutMainLoop();
	
	return 0;
}
